<?php

namespace Drupal\commerce_hubspot\EntitySync\Api;

use Drupal\commerce_hubspot\HubSpot\Helper as HubSpotHelper;
use Drupal\entity_sync\Client\ClientInterface as EntitySyncClientInterface;
use Drupal\hubspot_api\Manager as HubSpotManager;

/**
 * Entity Synchronization client adapter for the CRM Associations API.
 */
class AssociationClient implements EntitySyncClientInterface {

  /**
   * The HubSpot API manager.
   *
   * @var \Drupal\hubspot_api\Manager
   */
  protected $hubSpot;

  /**
   * The type of the origin object of the association.
   *
   * @var string
   */
  protected $fromObjectType;

  /**
   * The type of the target object of the association.
   *
   * @var string
   */
  protected $toObjectType;

  /**
   * The ID of the association type.
   *
   * @var int
   */
  protected $associationType;

  /**
   * The association category.
   *
   * @var string
   *
   * @see \HubSpot\Client\Crm\Associations\V4\Model\AssociationSpec::getAssociationCategoryAllowableValues()
   */
  protected $associationCategory;

  /**
   * Constructs a new AssociationClient object.
   *
   * @param \Drupal\hubspot_api\Manager $hubspot_manager
   *   The HubSpot API manager.
   * @param string $from_object_type
   *   The type of the origin object of the association.
   * @param string $to_object_type
   *   The type of the target object of the association.
   * @param int $association_type
   *   The ID of the association type.
   * @param int|null $association_category
   *   The association category; currently only used if using version 10+ of the
   *   HubSpot PHP SDK. It seems to be required in that case, otherwise an
   *   "Invalid input JSON" error is returned. We don't make it mandatory here
   *   though so that we maintain backwards compatibility.
   */
  public function __construct(
    HubSpotManager $hubspot_manager,
    string $from_object_type,
    string $to_object_type,
    int $association_type,
    string $association_category = NULL
  ) {
    $this->hubSpot = $hubspot_manager;
    $this->fromObjectType = $from_object_type;
    $this->toObjectType = $to_object_type;
    $this->associationType = $association_type;
    $this->associationCategory = $association_category;
  }

  /**
   * {@inheritdoc}
   */
  public function importList(array $filters = [], array $options = []) {
    throw new \Exception(
      'Importing a list of HubSpot CRM Associations is not supported.'
    );
  }

  /**
   * {@inheritdoc}
   */
  public function importEntity($id, array $options = []) {
    throw new \Exception(
      'Importing a HubSpot CRM Association is not supported yet.'
    );
  }

  /**
   * {@inheritdoc}
   */
  public function create(array $fields, array $options = []) {
    throw new \Exception(
      'CRM associations are currently considered updates as far the Entity Sync Client API is concerned because the entities involved in Drupal are mapped to their remote entities through different synchronizations and we do not need to track a remote ID that would be involved in a `create` action. That might change when we use separate entities for describing associations in Drupal, such as having User being group members of Organizations.'
    );
  }

  /**
   * {@inheritdoc}
   */
  public function update($id, array $fields, array $options = []) {
    $new_api = FALSE;
    $api = $this->hubSpot->getHandler()->crm();

    $client_method = HubSpotHelper::objectTypeToClientMethod(
      $this->fromObjectType
    );
    try {
      $api = $api->{$client_method}()->associationsApi();
    }
    // Version 10 of the PHP SDK has removed the object-specific associations
    // API client e.g. `deals()->associations()`. However, we don't want to drop
    // support for older versions yet because due to the dependency graph we'd
    // have to drop support for older versions of Drupal. We therefore try to
    // detect whether the old associations API exists, and if not use the new
    // one.
    // Unfortunately, there is no error class used specifically for discovery
    // errors nor do we have an easy way to detect the library version, so we
    // have to look for the error we're looking for from its message.
    catch (\Exception $exception) {
      $api = $api->associations()->v4()->basicApi();
      if ($exception->getMessage() !== 'Unable to discover "associationsApi" client') {
        throw $exception;
      }
      $new_api = TRUE;
    }

    if ($new_api === FALSE) {
      return $api->create(
        $id,
        $this->toObjectType,
        $fields['referenced_entity_remote_id'],
        $this->associationType
      )
      ->jsonSerialize();
    }

    $spec = new \HubSpot\Client\Crm\Associations\V4\Model\AssociationSpec([
      'association_category' => $this->associationCategory,
      'association_type_id' => $this->associationType,
    ]);
    return $api->create(
      $this->fromObjectType,
      $id,
      $this->toObjectType,
      $fields['referenced_entity_remote_id'],
      [$spec]
    )
    ->jsonSerialize();
  }

}
